! dataset.f90 -- a data type for storing simulated datasets
!
! Copyright (C) 2009-2010 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

module dataset
  use osl_core, wp => osl_wp
  implicit none

  type, public :: dataset_t
     integer :: n
     integer :: t
     integer :: dim_x
     real(wp), dimension(:,:), pointer :: y
     real(wp), dimension(:,:,:), pointer :: x
  end type dataset_t

contains

  ! Allocate memory for storing a dataset with the given dimensions.
  subroutine dataset_init(self, n, t, dim_x)
    type(dataset_t), intent(inout) :: self
    integer, intent(in) :: n
    integer, intent(in) :: t
    integer, intent(in) :: dim_x

    self%n = n
    self%t = t
    self%dim_x = dim_x
    allocate(self%y(t,n))
    allocate(self%x(dim_x,t,n))
  end subroutine dataset_init

  ! Free all memory associated with the given dataset.
  subroutine dataset_free(self)
    type(dataset_t), intent(inout) :: self
    deallocate(self%x)
    deallocate(self%y)
  end subroutine dataset_free

  ! Write the given dataset to the specified file.
  subroutine dataset_save(self, filename)
    type(dataset_t), intent(inout) :: self
    character(len=*), intent(in) :: filename
    integer :: ii, it

    open(UNIT=25, FILE=filename)
    do ii = 1, self%n
       do it = 1, self%t
          write(UNIT=25, FMT='(2i5,100g17.5)') &
               it, ii, self%y(it, ii), self%x(:,it,ii)
       end do
    end do
    close(UNIT=25)
  end subroutine dataset_save

end module dataset
