! osl_special.f90 -- special functions
!
! Copyright (C) 2005-2010 Jason R. Blevins
! All rights reserved.
!
! This file is part of the Open Scientific Library.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

module osl_special
  use osl_core, wp => osl_wp
  implicit none

  private

  public :: OSL_FACTORIAL_MAX

  public :: osl_gamma
  public :: osl_log_gamma
  public :: osl_factorial
  public :: osl_choose

  ! Maximum n such that n! does not overflow
  integer, parameter :: OSL_FACTORIAL_MAX = 170

  ! Lookup table for osl_factorial
  real(wp), dimension(OSL_FACTORIAL_MAX) :: osl_factorial_table
  data osl_factorial_table(1) / 1.0_wp /
  data osl_factorial_table(2) / 2.0_wp /
  data osl_factorial_table(3) / 6.0_wp /
  data osl_factorial_table(4) / 24.0_wp /
  data osl_factorial_table(5) / 120.0_wp /
  data osl_factorial_table(6) / 720.0_wp /
  data osl_factorial_table(7) / 5040.0_wp /
  data osl_factorial_table(8) / 40320.0_wp /
  data osl_factorial_table(9) / 362880.0_wp /
  data osl_factorial_table(10) / 3628800.0_wp /
  data osl_factorial_table(11) / 39916800.0_wp /
  data osl_factorial_table(12) / 479001600.0_wp /
  data osl_factorial_table(13) / 6227020800.0_wp /
  data osl_factorial_table(14) / 87178291200.0_wp /
  data osl_factorial_table(15) / 1307674368000.0_wp /
  data osl_factorial_table(16) / 20922789888000.0_wp /
  data osl_factorial_table(17) / 355687428096000.0_wp /
  data osl_factorial_table(18) / 6402373705728000.0_wp /
  data osl_factorial_table(19) / 121645100408832000.0_wp /
  data osl_factorial_table(20) / 2432902008176640000.0_wp /
  data osl_factorial_table(21) / 51090942171709440000.0_wp /
  data osl_factorial_table(22) / 1124000727777607680000.0_wp /
  data osl_factorial_table(23) / 25852016738884976640000.0_wp /
  data osl_factorial_table(24) / 620448401733239439360000.0_wp /
  data osl_factorial_table(25) / 15511210043330985984000000.0_wp /
  data osl_factorial_table(26) / 403291461126605635584000000.0_wp /
  data osl_factorial_table(27) / 10888869450418352160768000000.0_wp /
  data osl_factorial_table(28) / 304888344611713860501504000000.0_wp /
  data osl_factorial_table(29) / 8841761993739701954543616000000.0_wp /
  data osl_factorial_table(30) / 265252859812191058636308480000000.0_wp /
  data osl_factorial_table(31) / 8222838654177922817725562880000000.0_wp /
  data osl_factorial_table(32) / 263130836933693530167218012160000000.0_wp /
  data osl_factorial_table(33) / 8683317618811886495518194401280000000.0_wp /
  data osl_factorial_table(34) / 2.95232799039604140847618609644e38_wp /
  data osl_factorial_table(35) / 1.03331479663861449296666513375e40_wp /
  data osl_factorial_table(36) / 3.71993326789901217467999448151e41_wp /
  data osl_factorial_table(37) / 1.37637530912263450463159795816e43_wp /
  data osl_factorial_table(38) / 5.23022617466601111760007224100e44_wp /
  data osl_factorial_table(39) / 2.03978820811974433586402817399e46_wp /
  data osl_factorial_table(40) / 8.15915283247897734345611269600e47_wp /
  data osl_factorial_table(41) / 3.34525266131638071081700620534e49_wp /
  data osl_factorial_table(42) / 1.40500611775287989854314260624e51_wp /
  data osl_factorial_table(43) / 6.04152630633738356373551320685e52_wp /
  data osl_factorial_table(44) / 2.65827157478844876804362581101e54_wp /
  data osl_factorial_table(45) / 1.19622220865480194561963161496e56_wp /
  data osl_factorial_table(46) / 5.50262215981208894985030542880e57_wp /
  data osl_factorial_table(47) / 2.58623241511168180642964355154e59_wp /
  data osl_factorial_table(48) / 1.24139155925360726708622890474e61_wp /
  data osl_factorial_table(49) / 6.08281864034267560872252163321e62_wp /
  data osl_factorial_table(50) / 3.04140932017133780436126081661e64_wp /
  data osl_factorial_table(51) / 1.55111875328738228022424301647e66_wp /
  data osl_factorial_table(52) / 8.06581751709438785716606368564e67_wp /
  data osl_factorial_table(53) / 4.27488328406002556429801375339e69_wp /
  data osl_factorial_table(54) / 2.30843697339241380472092742683e71_wp /
  data osl_factorial_table(55) / 1.26964033536582759259651008476e73_wp /
  data osl_factorial_table(56) / 7.10998587804863451854045647464e74_wp /
  data osl_factorial_table(57) / 4.05269195048772167556806019054e76_wp /
  data osl_factorial_table(58) / 2.35056133128287857182947491052e78_wp /
  data osl_factorial_table(59) / 1.38683118545689835737939019720e80_wp /
  data osl_factorial_table(60) / 8.32098711274139014427634118320e81_wp /
  data osl_factorial_table(61) / 5.07580213877224798800856812177e83_wp /
  data osl_factorial_table(62) / 3.14699732603879375256531223550e85_wp /
  data osl_factorial_table(63) / 1.982608315404440064116146708360e87_wp /
  data osl_factorial_table(64) / 1.268869321858841641034333893350e89_wp /
  data osl_factorial_table(65) / 8.247650592082470666723170306800e90_wp /
  data osl_factorial_table(66) / 5.443449390774430640037292402480e92_wp /
  data osl_factorial_table(67) / 3.647111091818868528824985909660e94_wp /
  data osl_factorial_table(68) / 2.480035542436830599600990418570e96_wp /
  data osl_factorial_table(69) / 1.711224524281413113724683388810e98_wp /
  data osl_factorial_table(70) / 1.197857166996989179607278372170e100_wp /
  data osl_factorial_table(71) / 8.504785885678623175211676442400e101_wp /
  data osl_factorial_table(72) / 6.123445837688608686152407038530e103_wp /
  data osl_factorial_table(73) / 4.470115461512684340891257138130e105_wp /
  data osl_factorial_table(74) / 3.307885441519386412259530282210e107_wp /
  data osl_factorial_table(75) / 2.480914081139539809194647711660e109_wp /
  data osl_factorial_table(76) / 1.885494701666050254987932260860e111_wp /
  data osl_factorial_table(77) / 1.451830920282858696340707840860e113_wp /
  data osl_factorial_table(78) / 1.132428117820629783145752115870e115_wp /
  data osl_factorial_table(79) / 8.946182130782975286851441715400e116_wp /
  data osl_factorial_table(80) / 7.156945704626380229481153372320e118_wp /
  data osl_factorial_table(81) / 5.797126020747367985879734231580e120_wp /
  data osl_factorial_table(82) / 4.753643337012841748421382069890e122_wp /
  data osl_factorial_table(83) / 3.945523969720658651189747118010e124_wp /
  data osl_factorial_table(84) / 3.314240134565353266999387579130e126_wp /
  data osl_factorial_table(85) / 2.817104114380550276949479442260e128_wp /
  data osl_factorial_table(86) / 2.422709538367273238176552320340e130_wp /
  data osl_factorial_table(87) / 2.107757298379527717213600518700e132_wp /
  data osl_factorial_table(88) / 1.854826422573984391147968456460e134_wp /
  data osl_factorial_table(89) / 1.650795516090846108121691926250e136_wp /
  data osl_factorial_table(90) / 1.485715964481761497309522733620e138_wp /
  data osl_factorial_table(91) / 1.352001527678402962551665687590e140_wp /
  data osl_factorial_table(92) / 1.243841405464130725547532432590e142_wp /
  data osl_factorial_table(93) / 1.156772507081641574759205162310e144_wp /
  data osl_factorial_table(94) / 1.087366156656743080273652852570e146_wp /
  data osl_factorial_table(95) / 1.032997848823905926259970209940e148_wp /
  data osl_factorial_table(96) / 9.916779348709496892095714015400e149_wp /
  data osl_factorial_table(97) / 9.619275968248211985332842594960e151_wp /
  data osl_factorial_table(98) / 9.426890448883247745626185743100e153_wp /
  data osl_factorial_table(99) / 9.332621544394415268169923885600e155_wp /
  data osl_factorial_table(100) / 9.33262154439441526816992388563e157_wp /
  data osl_factorial_table(101) / 9.42594775983835942085162312450e159_wp /
  data osl_factorial_table(102) / 9.61446671503512660926865558700e161_wp /
  data osl_factorial_table(103) / 9.90290071648618040754671525458e163_wp /
  data osl_factorial_table(104) / 1.02990167451456276238485838648e166_wp /
  data osl_factorial_table(105) / 1.08139675824029090050410130580e168_wp /
  data osl_factorial_table(106) / 1.146280563734708354534347384148e170_wp /
  data osl_factorial_table(107) / 1.226520203196137939351751701040e172_wp /
  data osl_factorial_table(108) / 1.324641819451828974499891837120e174_wp /
  data osl_factorial_table(109) / 1.443859583202493582204882102460e176_wp /
  data osl_factorial_table(110) / 1.588245541522742940425370312710e178_wp /
  data osl_factorial_table(111) / 1.762952551090244663872161047110e180_wp /
  data osl_factorial_table(112) / 1.974506857221074023536820372760e182_wp /
  data osl_factorial_table(113) / 2.231192748659813646596607021220e184_wp /
  data osl_factorial_table(114) / 2.543559733472187557120132004190e186_wp /
  data osl_factorial_table(115) / 2.925093693493015690688151804820e188_wp /
  data osl_factorial_table(116) / 3.393108684451898201198256093590e190_wp /
  data osl_factorial_table(117) / 3.96993716080872089540195962950e192_wp /
  data osl_factorial_table(118) / 4.68452584975429065657431236281e194_wp /
  data osl_factorial_table(119) / 5.57458576120760588132343171174e196_wp /
  data osl_factorial_table(120) / 6.68950291344912705758811805409e198_wp /
  data osl_factorial_table(121) / 8.09429852527344373968162284545e200_wp /
  data osl_factorial_table(122) / 9.87504420083360136241157987140e202_wp /
  data osl_factorial_table(123) / 1.21463043670253296757662432419e205_wp /
  data osl_factorial_table(124) / 1.50614174151114087979501416199e207_wp /
  data osl_factorial_table(125) / 1.88267717688892609974376770249e209_wp /
  data osl_factorial_table(126) / 2.37217324288004688567714730514e211_wp /
  data osl_factorial_table(127) / 3.01266001845765954480997707753e213_wp /
  data osl_factorial_table(128) / 3.85620482362580421735677065923e215_wp /
  data osl_factorial_table(129) / 4.97450422247728744039023415041e217_wp /
  data osl_factorial_table(130) / 6.46685548922047367250730439554e219_wp /
  data osl_factorial_table(131) / 8.47158069087882051098456875820e221_wp /
  data osl_factorial_table(132) / 1.11824865119600430744996307608e224_wp /
  data osl_factorial_table(133) / 1.48727070609068572890845089118e226_wp /
  data osl_factorial_table(134) / 1.99294274616151887673732419418e228_wp /
  data osl_factorial_table(135) / 2.69047270731805048359538766215e230_wp /
  data osl_factorial_table(136) / 3.65904288195254865768972722052e232_wp /
  data osl_factorial_table(137) / 5.01288874827499166103492629211e234_wp /
  data osl_factorial_table(138) / 6.91778647261948849222819828311e236_wp /
  data osl_factorial_table(139) / 9.61572319694108900419719561353e238_wp /
  data osl_factorial_table(140) / 1.34620124757175246058760738589e241_wp /
  data osl_factorial_table(141) / 1.89814375907617096942852641411e243_wp /
  data osl_factorial_table(142) / 2.69536413788816277658850750804e245_wp /
  data osl_factorial_table(143) / 3.85437071718007277052156573649e247_wp /
  data osl_factorial_table(144) / 5.55029383273930478955105466055e249_wp /
  data osl_factorial_table(145) / 8.04792605747199194484902925780e251_wp /
  data osl_factorial_table(146) / 1.17499720439091082394795827164e254_wp /
  data osl_factorial_table(147) / 1.72724589045463891120349865931e256_wp /
  data osl_factorial_table(148) / 2.55632391787286558858117801578e258_wp /
  data osl_factorial_table(149) / 3.80892263763056972698595524351e260_wp /
  data osl_factorial_table(150) / 5.71338395644585459047893286526e262_wp /
  data osl_factorial_table(151) / 8.62720977423324043162318862650e264_wp /
  data osl_factorial_table(152) / 1.31133588568345254560672467123e267_wp /
  data osl_factorial_table(153) / 2.00634390509568239477828874699e269_wp /
  data osl_factorial_table(154) / 3.08976961384735088795856467036e271_wp /
  data osl_factorial_table(155) / 4.78914290146339387633577523906e273_wp /
  data osl_factorial_table(156) / 7.47106292628289444708380937294e275_wp /
  data osl_factorial_table(157) / 1.17295687942641442819215807155e278_wp /
  data osl_factorial_table(158) / 1.85327186949373479654360975305e280_wp /
  data osl_factorial_table(159) / 2.94670227249503832650433950735e282_wp /
  data osl_factorial_table(160) / 4.71472363599206132240694321176e284_wp /
  data osl_factorial_table(161) / 7.59070505394721872907517857094e286_wp /
  data osl_factorial_table(162) / 1.22969421873944943411017892849e289_wp /
  data osl_factorial_table(163) / 2.00440157654530257759959165344e291_wp /
  data osl_factorial_table(164) / 3.28721858553429622726333031164e293_wp /
  data osl_factorial_table(165) / 5.42391066613158877498449501421e295_wp /
  data osl_factorial_table(166) / 9.00369170577843736647426172359e297_wp /
  data osl_factorial_table(167) / 1.50361651486499904020120170784e300_wp /
  data osl_factorial_table(168) / 2.52607574497319838753801886917e302_wp /
  data osl_factorial_table(169) / 4.26906800900470527493925188890e304_wp /
  data osl_factorial_table(170) / 7.25741561530799896739672821113e306_wp /

contains

  !---------------------------------------------------------------------------!
  ! Gamma Function                                                            !
  !---------------------------------------------------------------------------!

  !------------------------------------!
  ! Gamma Function                     !
  !------------------------------------!

  ! Calculates the GAMMA function evaluated at a real value X.
  !
  ! This function is based on the Fortran 77 version from TOMS 715,
  ! dated March 12, 1992, by W. J. Cody and L. Stoltz, Argonne
  ! National Laboratory.  Computation is based on the algorithm
  ! outlined by Cody (1975) The program uses rational functions that
  ! approximate the GAMMA function to at least 20 significant decimal
  ! digits.  Coefficients for the approximation over the interval
  ! (1,2) are unpublished.  Those for the approximation for X >= 12
  ! are from Hart et al. (1968).  The accuracy achieved depends on the
  ! arithmetic system, the compiler, the intrinsic functions, and
  ! proper selection of the machine-dependent constants.
  !
  ! The following machine-dependent constants must be declared
  ! (IEEE values are provided as a default):
  !
  ! XBIG   - the largest argument for which GAMMA(X) is representable.
  ! XMININ - the smallest positive floating-point number such that
  !          1/XMININ is machine representable.
  !
  ! Error returns:
  !
  ! The program returns the value XINF for singularities or when
  ! overflow would occur.  The computation is believed to be free
  ! of underflow and overflow.
  !
  ! Last modified: August 27, 2010.
  !
  ! References:
  !
  !     Cody, W. J. (1976).  An Overview of Software Development for
  !     Special Functions, Lecture Notes in Mathematics 506, Numerical
  !     Analysis, ed. G. A. Watson. Springer Verlag, Berlin.
  !
  !     Hart, J. F., E. W. Cheney, C. L. Lawson, H. J. Maehly,
  !     C. K. Mesztenyi, J. R. Rice, H. C. Thacher Jr., and
  !     C. Witzgall (1968). Computer Approximations. Wiley and Sons,
  !     New York.
  !
  ! Changes from the original gamma.f:
  !
  !   * Converted to Fortran 90 for additional clarity.
  !   * Factor out CONV statement function in favor of REAL intrinsic.
  !   * Use OSL mathematical constants.
  !   * Use HUGE and EPSILON intrinsics for calculating XINF and EPS.
  !   * Rename SQRTPI constant to LOG2PI_2.
  !   * Rewrite conditionals to avoid use of GO TO.
  function osl_gamma(x) result(res)
    real(wp), intent(in) :: x
    real(wp) :: res

    ! Machine dependent parameters
    real(wp), parameter :: XBIG = 171.624_wp
    real(wp), parameter :: XMININ = 2.23e-308_wp

    ! Numeric parameters
    real(wp), parameter :: EPS = epsilon(1.0_wp)
    real(wp), parameter :: XINF = huge(1.0_wp)

    ! Mathematical constants
    real(wp), parameter :: ZERO = 0.0_wp
    real(wp), parameter :: HALF = 0.5_wp
    real(wp), parameter :: ONE = 1.0_wp
    real(wp), parameter :: TWO = 2.0_wp
    real(wp), parameter :: TWELVE = 12.0_wp
    real(wp), parameter :: LOG2PI_2 = 0.9189385332046727417803297_wp

    ! Local variables
    integer :: i, n
    logical :: parity
    real(wp) :: fact, sum, xden, xnum, y, y1, ysq, z

    ! Numerator and denominator coefficients for rational minimax
    ! approximation over (1, 2).
    real(wp), dimension(8) :: p = (/ &
         -1.71618513886549492533811e+0_wp,  2.47656508055759199108314e+1_wp, &
         -3.79804256470945635097577e+2_wp,  6.29331155312818442661052e+2_wp, &
          8.66966202790413211295064e+2_wp, -3.14512729688483675254357e+4_wp, &
         -3.61444134186911729807069e+4_wp,  6.64561438202405440627855e+4_wp /)

    real(wp), dimension(8) :: q = (/ &
         -3.08402300119738975254353e+1_wp,  3.15350626979604161529144e+2_wp, &
         -1.01515636749021914166146e+3_wp, -3.10777167157231109440444e+3_wp, &
          2.25381184209801510330112e+4_wp,  4.75584627752788110767815e+3_wp, &
         -1.34659959864969306392456e+5_wp, -1.15132259675553483497211e+5_wp /)

    ! Coefficients for minimax approximation over (12, INF).
    real(wp), dimension(7) :: c = (/ &
         -1.910444077728e-03_wp, 8.4171387781295e-04_wp, &
         -5.952379913043012e-04_wp, 7.93650793500350248e-04_wp, &
         -2.777777777777681622553e-03_wp, 8.333333333333333331554247e-02_wp, &
          5.7083835261e-03_wp /)

    parity = .false.
    fact = ONE
    n = 0
    y = x
    if (y <= ZERO) then
       ! Argument is negative
       y = -x
       y1 = aint(y)
       res = y - y1
       if (res /= ZERO) then
          if (y1 /= aint(y1*HALF)*TWO) then
             parity = .true.
          end if
          fact = -OSL_PI / sin(OSL_PI * res)
          y = y + ONE
       else
          res = XINF
          return
       end if
    end if

    if (y < eps) then
       ! Case: 0.0 < argument < EPS
       if (y >= xminin) then
          res = ONE / y
       else
          res = xinf
          return
       end if
    else if (y < TWELVE) then
       y1 = y
       if (y < ONE) then
          ! EPS < argument < 1.0
          z = y
          y = y + ONE
       else
          ! Case: 1.0 <= argument <= 12.0, reduce argument if necessary
          n = int(y) - 1
          y = y - real(n, wp)
          z = y - ONE
       end if
       ! Evaluate approximation for 1.0 < argument < 2.0
       xnum = ZERO
       xden = ONE
       do i = 1, 8
          xnum = (xnum + p(i)) * z
          xden = xden * z + q(i)
       end do
       res = xnum / xden + ONE
       if (y1 < y) then
          ! Adjust result for case: 0.0 < argument < 1.0
          res = res / y1
       else if (y1 > y) then
          ! Adjust result for case: 2.0 < argument < 12.0
          do i = 1, n
             res = res * y
             y = y + ONE
          end do
       end if
    else
       ! Evaluate for argument >= 12.0
       if (y <= XBIG) then
          ysq = y * y
          sum = c(7)
          do i = 1, 6
             sum = sum / ysq + c(i)
          end do
          sum = sum/y - y + LOG2PI_2
          sum = sum + (y-half) * log(y)
          res = exp(sum)
       else
          res = XINF
          return
       end if
    end if

    ! Final adjustments and return
    if (parity) then
       res = -res
    end if
    if (fact /= one) then
       res = fact / res
    end if
  end function osl_gamma

  !------------------------------------!
  ! Log Gamma Function                 !
  !------------------------------------!

  ! Calculates the logarithm of the gamma function using the algorithm
  ! of Macleod (1989).
  !
  ! Reference:
  !
  !     Macleod, Allan (1989).  Algorithm AS 245: A Robust and
  !     Reliable Algorithm for the Logarithm of the Gamma Function.
  !     Applied Statistics 38(2), 397-402.
  function osl_log_gamma(x, status) result(f)
    real(wp), intent(in) :: x
    integer, intent(out), optional :: status
    real(wp) :: f

    ! Machine-dependant constants for IEEE double precision
    ! (B = 2, t = 53 and U = 1023).  See p. 399 for details.
    real(wp), parameter :: XLGE = 5.10e6_wp
    real(wp), parameter :: XLGST = huge(1.0_wp)

    ! Coefficients of rational functions
    real(wp), parameter :: r1(9) = (/ &
          -2.66685511495_wp, -24.4387534237_wp, -21.9698958928_wp, &
          11.1667541262_wp,    3.13060547623_wp,  0.607771387771_wp, &
          11.9400905721_wp,   31.4690115749_wp,  15.2346874070_wp /)
    real(wp), parameter :: r2(9) = (/ &
         -78.3359299449_wp, -142.046296688_wp, 137.519416416_wp, &
         78.6994924154_wp, 4.16438922228_wp,  47.0668766060_wp, &
         313.399215894_wp,  263.505074721_wp, 43.3400022514_wp /)
    real(wp), parameter :: r3(9) = (/ &
         -2.12159572323E5_wp, 2.30661510616E5_wp, 2.74647644705E4_wp, &
         -4.02621119975E4_wp, -2.29660729780E3_wp, -1.16328495004E5_wp, &
         -1.46025937511E5_wp, -2.42357409629E4_wp, -5.70691009324E2_wp /)
    real(wp), parameter :: r4(5) = (/ &
         0.279195317918525_wp, 0.4917317610505968_wp, 0.0692910599291889_wp, &
         3.350343815022304_wp, 6.012459259764103_wp /)

    ! Fixed constants
    real(wp), parameter :: ALR2PI = 0.918938533204673_wp, &
         FOUR = 4._wp, HALF = 0.5_wp, ONE = 1.0_wp, ONEP5 = 1.5_wp, &
         TWELVE = 12.0_wp, ZERO = 0.0_wp

    ! Local variables
    real(wp) :: z, z1, z2, y

    z = x
    f = ZERO

    ! Test for valid function argument
    if (z >= XLGST) then
       call osl_status(OSL_EDOM, status, 'X too large in OSL_LOG_GAMMA')
       return
    end if
    if (z <= ZERO) then
       call osl_status(OSL_EDOM, status, 'X must be positive in OSL_LOG_GAMMA')
       return
    end if
    call osl_status(OSL_SUCCESS, status)

    ! Calculation for 0 < Z < 0.5 and 0.5 <= Z < 1.5 combined
    if (z < ONEP5) then
       IF (z < HALF) then
          f = -log(z)
          y = z + ONE
    
          ! Test whether Z < machine epsilon
          if (y == ONE) then
             return
          end if
       else
          f = ZERO
          y = z
          z = (z - HALF) - HALF
       end if
       f = f + z*((((r1(5)*y + r1(4))*y + r1(3))*y + r1(2))*y + r1(1))&
              / ((((y + r1(9))*y + r1(8))*y+ r1(7))*y + r1(6))
       return
    end if

    ! Calculation for 1.5 <= Z < 4.0
    if (z < FOUR) then
       y = (z - ONE) - ONE
       f = y * ((((r2(5)*z + r2(4))*z + r2(3))*z + r2(2))*z + r2(1)) &
              / ((((z + r2(9))*z + r2(8))*z + r2(7))*z+ r2(6))
       return
    end if

    ! Calculation for 4.0 <= Z < 12.0
    if (z < twelve) then
       f = ((((r3(5)*z + r3(4))*z + r3(3))*z + r3(2))*z + r3(1)) &
              / ((((z + r3(9))*z + r3(8))*z + r3(7))*z + r3(6))
       return
    end if

    ! Calculation for Z >= 12.0
    y = log(z)
    f = z * (y - ONE) - HALF * y + ALR2PI
    if (z > XLGE) then
       return
    end if
    z1 = ONE / z
    z2 = z1 * z1
    f = f + z1 * ((r4(3)*z2 + r4(2))*z2 + r4(1)) / ((z2 + r4(5))*z2 + r4(4))
    return
  end function osl_log_gamma

  !--------------------------------------------------------------------------!
  ! Factorial and Related Functions                                          !
  !--------------------------------------------------------------------------!

  !------------------------------------!
  ! Factorial                          !
  !------------------------------------!

  ! Returns N!, equal to the product of the integers 1, 2, ..., N.
  !
  ! The return value is exact for values of N smaller than 18.
  ! For larger values of N, the error is on the order of
  ! 2 * EPSILON(1.0_WP) * N!.
  function osl_factorial(n, err, status)
    integer, intent(in) :: n
    real(wp), intent(out), optional :: err
    integer, intent(out), optional :: status
    real(wp) :: osl_factorial
    integer :: code
    real(wp) :: error

    error = 0.0_wp
    osl_factorial = 0.0_wp
    code = OSL_SUCCESS

    if (n < 0) then
       code = OSL_EDOM  ! domain is N >= 0
    else if (n == 0) then
       osl_factorial = 1.0_wp
    else if (n < 18) then
       osl_factorial = osl_factorial_table(n)
    else if (n <= osl_factorial_max) then
       osl_factorial = osl_factorial_table(n)
       error = 2.0_wp * epsilon(1.0_wp) * osl_factorial
    else
       code = OSL_EOVRFLW  ! would result in overflow
    end if

    if (present(err)) then
       err = error
    end if

    if (present(status)) then
       status = code
    end if
  end function osl_factorial

  !------------------------------------!
  ! Binomial Coefficients              !
  !------------------------------------!

  ! Return the number of combinations of K elements chosen from a
  ! collection of N elements.
  function osl_choose(n, k)
    integer, intent(in) :: n
    integer, intent(in) :: k
    integer :: osl_choose
    integer :: i, kk
    real(wp) :: c

    ! handle invalid values
    if (k > n .or. k < 0) then
       osl_choose = 0
       return
    end if

    ! use symmetry
    if (k > n / 2) then
       kk = n - k
    else
       kk = k
    end if

    ! use the definition of the binomial coefficient
    c = 1.0_wp
    do i = 1, kk
       c = c * real(n - kk + i, wp) / real(i, wp)
    end do

    ! avoid rounding error
    osl_choose = int(real(c, wp) + 0.5_wp)
  end function osl_choose

  !--------------------------------------------------------------------------!
  ! Sigmoid Function                                                         !
  !--------------------------------------------------------------------------!

  ! Takes as input a vector of length N-1 and passes it through a
  ! multidimensional sigmoid mapping to return a vector of length N whose
  ! elements are nonnegative and sum to one.  The I-th element of the
  ! input vector controls the relative magnitude of the I-th element of
  ! the output vector.  The value corresponding to the N-th element is
  ! internally normalized to 1.  Thus, if input element J is smaller than 1,
  ! then on output element J will be smaller than element N.
  ! Similarly if input element J is larger than 1.
  !
  ! This function is useful for performing unconstrained optimization on
  ! parameters which correspond to a complete set of probabilities, since
  ! each of them must be nonnegative and the set must sum to one.
  function osl_sigmoid(theta)
    real(wp), dimension(:), intent(in) :: theta
    real(wp), dimension(size(theta,1)+1) :: osl_sigmoid
    real(wp), dimension(size(theta,1)) :: exptheta
    integer :: n

    n = size(theta,1) + 1
    exptheta = exp(theta)

    osl_sigmoid(1:n-1) = exptheta
    osl_sigmoid(n) = 1.0_wp
    osl_sigmoid = osl_sigmoid / (1.0_wp + sum(exptheta)) ! Normalize
  end function osl_sigmoid

end module osl_special
